from radicli import Arg
from wasabi import msg

from .. import ty
from ..auth import get_current_auth
from ..cli import cli
from ..errors import CLIError, ProdigyTeamsErrors
from ..messages import Messages
from ..prodigy_teams_pam_sdk.models import ProjectCreating, ProjectReturning
from ..query import resolve_project, resolve_project_id
from ..ui import print_info_table, print_table_with_select


@cli.subcommand(
    "projects",
    "list",
    # fmt: off
    select=Arg("--select", help=Messages.select.format(opts=list(ProjectReturning.__fields__))),
    name=Arg("--name", help=Messages.filter_by.format(filter="name")),
    # fmt: on
)
def list(
    select: ty.List[str] = ["id", "name"], name: ty.Optional[str] = None
) -> ty.Sequence[ProjectReturning]:
    """List all projects"""
    client = get_current_auth().client
    res = client.project.all(name=name)
    print_table_with_select(res.items, select=select)
    return res.items


@cli.subcommand(
    "projects",
    "info",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="project")),
    as_json=Arg("--json", help=Messages.as_json),
)
def info(name_or_id: ty.StrOrUUID, as_json: bool = False) -> ProjectReturning:
    """Get detailed info for a project"""
    res = resolve_project(name_or_id)
    print_info_table(res, as_json=as_json)
    return res


@cli.subcommand(
    "projects",
    "create",
    name=Arg(help=Messages.name.format(noun="project")),
    description=Arg(help=Messages.description.format(noun="project")),
    exists_ok=Arg("--exists-ok", help=Messages.exists_ok),
)
def create(
    name: str, description: str, exists_ok: bool = False
) -> ty.Optional[ty.UUID]:
    """Create a new project"""
    auth = get_current_auth()
    client = auth.client
    body = ProjectCreating(name=name, description=description)
    try:
        res = client.project.create(body)
    except ProdigyTeamsErrors.ProjectExists:
        if exists_ok:
            msg.info(Messages.T001.format(noun="project", name=name))
            return None
        raise CLIError(Messages.E002.format(noun="project", name=name))
    except ProdigyTeamsErrors.ProjectInvalid:
        raise CLIError(Messages.E004.format(noun="project", name=name))
    msg.divider("Project")
    msg.table(res.dict())
    return res.id


@cli.subcommand(
    "projects",
    "delete",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="project")),
)
def delete(name_or_id: ty.StrOrUUID) -> ty.UUID:
    """Delete a project"""
    project_id = resolve_project_id(name_or_id)
    auth = get_current_auth()
    try:
        auth.client.project.delete(id=project_id)
    except (
        ProdigyTeamsErrors.ProjectForbiddenDelete,
        ProdigyTeamsErrors.ProjectNotFound,
    ):
        raise CLIError(Messages.E006.format(noun="project", name=name_or_id))
    else:
        msg.good(Messages.T003.format(noun="project", name=name_or_id))
    return project_id



from . import (
    actions,
    assets,
    clusters,
    config,
    datasets,
    files,
    general,
    import_export,
    packages,
    paths,
    projects,
    recipes,
    tasks,
)

__all__ = [
    "assets",
    "clusters",
    "config",
    "datasets",
    "files",
    "packages",
    "paths",
    "projects",
    "recipes",
    "actions",
    "tasks",
    "general",
    "import_export",
]



from ..models import (
    ObjectValidation,
    RecipePlanCreating,
    RecipePlanDeleting,
    RecipePlanReading,
    RecipePlanReturning,
    RecipePlanUpdating,
)
from .base import ModelClient


class RecipePlan(
    ModelClient[
        RecipePlanCreating,
        RecipePlanReading,
        RecipePlanUpdating,
        RecipePlanDeleting,
        RecipePlanReturning,
        RecipePlanReturning,
    ]
):
    Creating = RecipePlanCreating
    Reading = RecipePlanReading
    Updating = RecipePlanUpdating
    Deleting = RecipePlanDeleting
    Returning = RecipePlanReturning

    def validate_objects(self, data: RecipePlanCreating) -> ObjectValidation:
        res = self.request(
            "POST",
            endpoint="validate-objects",
            data=data,
            return_model=ObjectValidation,
        )
        assert isinstance(res, ObjectValidation)
        return res

    async def validate_objects_async(
        self, data: RecipePlanCreating
    ) -> ObjectValidation:
        res = await self.request_async(
            "POST",
            endpoint="validate-objects",
            data=data,
            return_model=ObjectValidation,
        )
        assert isinstance(res, ObjectValidation)
        return res



import sys

from radicli import (
    DEFAULT_PLACEHOLDER,
    Arg,
    ArgparseArg,
    ArgumentParser,
    ConverterType,
    get_arg,
)
from wasabi import msg

from .. import ty
from ..auth import CLIAuth, get_current_auth
from ..cli import CONVERTERS, cli
from ..config import SavedSettings
from ..errors import CLIError, ProdigyTeamsErrors
from ..messages import Messages
from ..prodigy_teams_broker_sdk.models import JobID, JobType
from ..prodigy_teams_pam_sdk import Client
from ..prodigy_teams_pam_sdk.models import (
    ActionCreating,
    ObjectValidation,
    RecipeDetail,
    RecipeListingLatest,
    RecipePlanCreating,
    TaskCreating,
)
from ..prodigy_teams_pam_sdk.recipe_utils import (
    Field,
    RecipeSchema,
    get_field_cli_args,
    get_field_optional,
    parse_plan,
)
from ..query import resolve_project_id
from ..ui import RecipeHelpFormatter, get_recipe_desc


def get_converter(arg_type: ty.Type) -> ty.Optional[ConverterType]:
    return CONVERTERS.get(arg_type, None)


def request_recipes(
    auth: CLIAuth, is_action: ty.Optional[bool] = None
) -> ty.Dict[str, RecipeDetail]:
    """Fetch all Recipes from Prodigy Teams
    auth (CLIAuth): CLI Auth object to handle auth to the Prodigy Teams API
    RETURNS (Tuple[Dict[str, RecipeDetail], Dict[str, RecipeDetail]]): Tuple of task_schemas, recipe_schemas
    """
    client = auth.client
    if client is None:
        return {}
    with msg.loading(Messages.T020):
        recipes = auth.retry(client.recipe.all_latest)(
            body=RecipeListingLatest(org_id=auth.org_id, broker_id=auth.broker_id)
        )
    schemas = {}
    for recipe in recipes.items:
        if is_action is None or recipe.is_action == is_action:
            schemas[recipe.name] = recipe
    return schemas


def create_from_recipe(
    recipe: RecipeDetail,
    args: ty.List[str],
    *,
    command: str,
    show_help: bool,
    exists_ok: bool = False,
) -> ty.Tuple[ty.UUID, RecipePlanCreating]:
    try:
        task_or_action_name, project_id, plan = _parse_plan(
            recipe,
            args,
            parser_prog=f"{cli.prog} {command} create {recipe.name.lower()}",
            show_help=show_help,
            get_project_id=_get_default_project_id,
        )
    except ShowHelp as e:
        print(e.help_text)
        sys.exit(1)
    auth = get_current_auth()
    object_validation = _get_object_validation(plan)
    if object_validation.errors:
        errors = ["\n".join(err) for err in object_validation.errors.values()]
        noun = "action" if recipe.is_action else "task"
        err = Messages.E044.format(noun=noun, name=task_or_action_name)
        raise CLIError(err, "\n".join(errors))
    _validate_plan(recipe, plan, object_validation)
    if recipe.is_action:
        job_id = _create_action(
            auth.client, task_or_action_name, project_id, plan, exists_ok=exists_ok
        )
    else:
        job_id = _create_task(
            auth.client, task_or_action_name, project_id, plan, exists_ok=exists_ok
        )
    return job_id.id, plan


def _parse_plan(
    recipe: RecipeDetail,
    args: ty.List[str],
    *,
    parser_prog: str,
    show_help: bool,
    get_project_id: ty.Callable[[], ty.Optional[ty.UUID]],
) -> ty.Tuple[str, ty.UUID, RecipePlanCreating]:
    # This function works hard to be pure and not rely on
    # external state. This way we'll be able to work with it in
    # the compatibility tests more easily.
    noun = "action" if recipe.is_action else "task"
    ap_args = _get_argparse_args(recipe)
    parser = _get_argparse_parser(parser_prog, recipe, ap_args)
    if not args or show_help:
        raise ShowHelp(parser.format_help())
    task_or_action_name, project_id, plan_args = _parse_cli_args(parser, args)
    missing = _check_missing(recipe, ap_args, plan_args)
    if missing:
        err = Messages.E001.format(noun=noun, name=task_or_action_name)
        raise CLIError(err, Messages.E043.format(args=", ".join(missing)))
    if project_id is None:
        project_id = get_project_id()
        if project_id is None:
            raise CLIError(Messages.E042.format(noun=noun, name=task_or_action_name))
    plan = RecipePlanCreating(
        recipe_id=recipe.id, project_id=project_id, args=plan_args
    )
    return task_or_action_name, project_id, plan


def _get_default_project_id() -> ty.Optional[ty.UUID]:
    project_id = SavedSettings.load().project
    if project_id is not None:
        project_id = resolve_project_id(project_id)
    return project_id


def _parse_cli_args(
    parser: ArgumentParser,
    args: ty.List[str],
) -> ty.Tuple[str, ty.Optional[ty.UUID], ty.Dict[str, Arg]]:
    parsed = vars(parser.parse_args(args))
    task_or_action_name = parsed.pop("name")
    assert isinstance(task_or_action_name, str)
    project_id = parsed.pop("project_id")
    if project_id is not None:
        project_id = ty.UUID(project_id)
    return (task_or_action_name, project_id, parsed)


def _check_missing(recipe: RecipeDetail, ap_args, parsed) -> ty.List[str]:
    optional = get_field_optional(recipe.form_schema)
    missing = []
    # We're iterating over the original args here because that gives us access
    # to the option names in the right CLI syntax
    for arg in ap_args:
        required = not optional.get(arg.id, True)
        if required and arg.id not in parsed:
            missing.append(arg.arg.option or arg.id)
    return missing


def _get_argparse_args(recipe: RecipeDetail) -> ty.List[ArgparseArg]:
    noun = "action" if recipe.is_action else "task"
    ap_args: ty.List[ArgparseArg] = []
    ap_args.append(get_arg("name", Arg(help=Messages.name.format(noun=noun)), str))
    for field in recipe.form_schema.fields:
        ap_args.extend(
            get_field_cli_args(
                ty.cast(Field, field), recipe.form_schema.cli_names or {}
            )
        )
    return ap_args


def _get_argparse_parser(
    prog: str, recipe: RecipeDetail, ap_args: ty.List[ArgparseArg]
) -> ArgumentParser:
    noun = "action" if recipe.is_action else "task"
    # To get argument parsing and consistent help docs out-of-the-box we create
    # a dummy parser with radicli's modified ArgumentParser (argparse subclass)
    p = ArgumentParser(
        prog=prog,
        description=get_recipe_desc(recipe),
        formatter_class=RecipeHelpFormatter,
    )
    for arg in ap_args:
        a_args, a_kwargs = arg.to_argparse()
        # We don't want defaults to creep in here, we want that to be
        # introduced at the last moment by the actual function.
        if "default" in a_kwargs:
            a_kwargs["default"] = DEFAULT_PLACEHOLDER
        p.add_argument(*a_args, **a_kwargs)
    # Add default arguments to all recipe CLIs
    pid_arg = Arg("--project-id", help=Messages.recipe_project_id.format(noun=noun))
    pid = get_arg("project_id", pid_arg, str, default=None)
    pid_args, pid_kwargs = pid.to_argparse()
    p.add_argument(*pid_args, **pid_kwargs)
    return p


def _get_object_validation(plan: RecipePlanCreating) -> ObjectValidation:
    auth = get_current_auth()
    object_validation = auth.client.recipeplan.validate_objects(plan)
    return object_validation


def _validate_plan(
    recipe: RecipeDetail,
    plan: RecipePlanCreating,
    object_validation: ObjectValidation,
) -> None:
    # Building a dummy objects map for parse_plan – the values don't matter,
    # since we're also only using a dummy fuction for create_custom_type
    objects_map = {"asset": {}, "dataset": {}}
    for field_type, objects in object_validation.existing.items():
        objects_map[field_type] = {arg: {} for arg in objects}
    for field_type, objects in object_validation.to_create.items():
        objects_map[field_type] = {arg: {} for arg in objects}
    parse_plan(
        ty.cast(RecipeSchema, recipe.form_schema),
        plan.args,
        objects_map,
        lambda x: lambda *a, **b: x,
    )


def _create_task(
    client: Client,
    name: str,
    project_id: ty.UUID,
    plan: RecipePlanCreating,
    exists_ok: bool = False,
) -> JobID:
    body = TaskCreating(name=name, project_id=project_id, plan=plan)
    try:
        task = client.task.create(body)
    except ProdigyTeamsErrors.ProjectNotFound:
        err = Messages.E013.format(name=project_id)
        raise CLIError(err)
    except ProdigyTeamsErrors.TaskExists:
        if exists_ok:
            msg.info(Messages.T001.format(noun="task", name=name))
            task = client.task.read(name=name, project_id=project_id)
            return JobID(id=task.id, job_type=JobType.task)
        err = Messages.E014.format(noun="task", name=name, project=project_id)
        raise CLIError(err)
    except ProdigyTeamsErrors.TaskInvalid:
        raise CLIError(Messages.E004.format(noun="task", name=name))
    return JobID(id=task.id, job_type=JobType.task)


def _create_action(
    client: Client,
    name: str,
    project_id: ty.UUID,
    plan: RecipePlanCreating,
    exists_ok: bool = False,
) -> JobID:
    body = ActionCreating(name=name, project_id=project_id, evaluation="", plan=plan)
    try:
        action = client.action.create(body)
    except ProdigyTeamsErrors.ProjectNotFound:
        err = Messages.E013.format(name=project_id)
        raise CLIError(err)
    except ProdigyTeamsErrors.ActionExists:
        if exists_ok:
            msg.info(Messages.T001.format(noun="action", name=name))
            action = client.action.read(name=name, project_id=project_id)
            return JobID(id=action.id, job_type=JobType.action)
        err = Messages.E014.format(noun="action", name=name, project=project_id)
        raise CLIError(err)
    except ProdigyTeamsErrors.ActionInvalid:
        raise CLIError(Messages.E004.format(noun="action", name=name))
    return JobID(id=action.id, job_type=JobType.action)


class ShowHelp(ValueError):
    def __init__(self, help_text: str):
        self.help_text = help_text



"""Script and module to interpret version.txt

N.B.

Make sure this file doesn't have any Python
dependencies and doesn't import from the rest of the library
at all. That way it can be invoked:

* As an import within the about.py of the different modules
* In a subprocess during the setup.py
* As a script in CI actions.

We want to make sure you can run like:

    python prodigy_teams_pam/versioning.py

N.B. 2:

This script is defined in shared/, and copied into the
different packages by:

    pdcli dev sync-shared

Don't modify the copies in the libraries.
"""
from dataclasses import dataclass
from pathlib import Path
from typing import Literal, Optional, Tuple


@dataclass
class Version:
    major: int
    minor: int
    patch: int

    @classmethod
    def from_string(cls, string: str):
        # It's a bit tricky to import packaging here, as we'd need to set it in
        # the setup_requires and stuff. We're not supporting alpha versions and
        # stuff anyway, so just assume we can parse it as three ints.
        # We need to interpret the version as ints, not a str, otherwise it won't
        # sort correctly.
        version = [int(piece) for piece in string.split(".")]
        while len(version) < 3:
            version.append(0)
        return cls(*version)

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"


@dataclass
class VersionIncrement:
    version: Version
    increment: Literal["", "major", "minor", "patch"]

    @classmethod
    def from_string(cls, string: str) -> "VersionIncrement":
        pieces = string.split("_")
        if len(pieces) != 4:
            raise ValueError(f"Cannot parse version increment: {string}")
        version_str, _, instruction, _2 = pieces
        assert instruction in ("", "major", "minor", "patch")
        return cls(Version.from_string(version_str), instruction)

    @property
    def sort_key(self) -> Tuple[int, int, int, int]:
        increment_priority = {"": 4, "major": 3, "minor": 2, "patch": 1}
        return (
            self.version.major,
            self.version.minor,
            self.version.patch,
            increment_priority[self.increment],
        )

    def compute(self, current: Optional[Version]) -> Version:
        if current is None:
            current = self.version
        if self.increment == "":
            return current
        elif self.increment == "major":
            return Version(current.major + 1, 0, 0)
        elif self.increment == "minor":
            return Version(current.major, current.minor + 1, 0)
        elif self.increment == "patch":
            return Version(current.major, current.minor, current.patch + 1)
        else:
            raise ValueError(f"Cannot parse version instruction: {self.increment}")


def read_version(path: Path) -> str:
    version_incs = [
        VersionIncrement.from_string(p.name)
        for p in path.iterdir()
        if p.suffix != ".py" and not p.is_dir()
    ]
    version_incs.sort(key=lambda v: v.sort_key)
    current_version = None
    for version_inc in version_incs:
        current_version = version_inc.compute(current_version)
    return str(current_version)


if __name__ == "__main__":
    import argparse

    P = argparse.ArgumentParser(
        description="Interpret a version.txt and print the version"
    )
    P.add_argument("path", type=Path, help="Path to the version.txt")
    A = P.parse_args()
    print(read_version(A.path))



from ..models import (
    PackageCreating,
    PackageDeleting,
    PackageListingLatest,
    PackageReading,
    PackageReadingLatest,
    PackageReturning,
    PackageUpdating,
)
from .base import VersionedModelClient


class Package(
    VersionedModelClient[
        PackageCreating,
        PackageReading,
        PackageUpdating,
        PackageDeleting,
        PackageReturning,
        PackageReturning,
        PackageReadingLatest,
        PackageListingLatest,
    ]
):
    Creating = PackageCreating
    Reading = PackageReading
    Updating = PackageUpdating
    Deleting = PackageDeleting
    Returning = PackageReturning
    Detail = PackageReturning
    ReadingLatest = PackageReadingLatest
    ListingLatest = PackageListingLatest



from pathlib import Path

import srsly
from radicli import Arg
from wasabi import msg

from .. import ty
from ..auth import get_current_auth
from ..cli import cli
from ..errors import CLIError, ProdigyTeamsErrors
from ..messages import Messages
from ..prodigy_teams_pam_sdk.models import DatasetDetail, DatasetReturning
from ..query import collect_from_pages, resolve_dataset, resolve_dataset_id
from ..ui import PageProgress, print_info_table, print_table_with_select


@cli.subcommand(
    "datasets",
    "list",
    # fmt: off
    select=Arg("--select", help=Messages.select.format(opts=list(DatasetReturning.__fields__))),
    # fmt: on
)
def list(
    select: ty.List[str] = ["id", "name", "kind"]
) -> ty.Sequence[DatasetReturning]:
    """List all Datasets"""
    client = get_current_auth().client
    res = client.dataset.all()
    print_table_with_select(res.items, select=select)
    return res.items


@cli.subcommand(
    "datasets",
    "info",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="dataset")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="dataset")),
    as_json=Arg("--json", help=Messages.as_json),
)
def info(
    name_or_id: ty.StrOrUUID,
    cluster_id: ty.Optional[ty.UUID] = None,
    as_json: bool = False,
) -> DatasetReturning:
    """Get detailed info for a Dataset"""
    res = resolve_dataset(name_or_id, broker_id=cluster_id)
    print_info_table(res, as_json=as_json)
    return res


@cli.subcommand(
    "datasets",
    "create",
    name=Arg(help=Messages.name.format(noun="dataset")),
    kind=Arg("--kind", help=Messages.kind.format(noun="dataset")),
    exists_ok=Arg("--exists-ok", help=Messages.exists_ok),
)
def create(
    name: str, kind: str, exists_ok: bool = False
) -> ty.Union[DatasetReturning, DatasetDetail, None]:
    """Create a new dataset"""
    auth = get_current_auth()
    client = auth.client
    broker_id = auth.broker_id
    try:
        res = client.dataset.create(name=name, kind=kind, broker_id=broker_id)
    except ProdigyTeamsErrors.DatasetExists:
        if exists_ok:
            msg.info(Messages.T001.format(noun="dataset", name=name))
            return None
        raise CLIError(Messages.E002.format(noun="dataset", name=name))
    except ProdigyTeamsErrors.DatasetInvalid:
        raise CLIError(Messages.E004.format(noun="dataset", name=name))
    except ProdigyTeamsErrors.DatasetForbiddenCreate:
        raise CLIError(Messages.E003.format(noun="dataset", name=name))
    msg.divider("Dataset")
    msg.table(res.dict())
    return res


@cli.subcommand(
    "datasets",
    "delete",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="dataset")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="dataset")),
)
def delete(
    name_or_id: ty.StrOrUUID,
    cluster_id: ty.Optional[ty.UUID] = None,
) -> ty.UUID:
    """Delete a dataset"""
    dataset_id = resolve_dataset_id(name_or_id, broker_id=cluster_id)
    auth = get_current_auth()
    try:
        auth.client.dataset.delete(id=dataset_id)
    except (
        ProdigyTeamsErrors.ProjectForbiddenDelete,
        ProdigyTeamsErrors.ProjectNotFound,
    ):
        raise CLIError(Messages.E006.format(noun="dataset", name=name_or_id))
    else:
        msg.good(Messages.T003.format(noun="dataset", name=name_or_id))
    return dataset_id


@cli.subcommand(
    "datasets",
    "export",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="dataset")),
    output_file=Arg("--output", "-o", help=Messages.export_output),
)
def export(
    name_or_id: ty.StrOrUUID,
    output_file: ty.Union[str, Path] = "-",
) -> None:
    """Export all the examples from a dataset and save it in the designated file as
    JSONL (newline-delimited JSON).
    """
    auth = get_current_auth()
    cluster_id = auth.broker_id

    dataset = resolve_dataset(name_or_id, broker_id=cluster_id)

    def get_examples_page(page: int):
        return auth.broker_client.datasets.read_examples(
            dataset.name,
            page=page,
        )

    with PageProgress(collect_from_pages(get_examples_page)) as pages:
        srsly.write_jsonl(
            output_file,
            (example.content for page in pages for example in page.items),
            append_new_line=False,
        )



from .. import ty
from ..models import (
    PersonCreating,
    PersonDeleting,
    PersonReading,
    PersonReturning,
    PersonUpdating,
)
from .base import ModelClient


class Person(
    ModelClient[
        PersonCreating,
        PersonReading,
        PersonUpdating,
        PersonDeleting,
        PersonReturning,
        PersonReturning,
    ]
):
    Creating = PersonCreating
    Reading = PersonReading
    Updating = PersonUpdating
    Deleting = PersonDeleting
    Returning = PersonReturning

    class Authenticating(ty.BaseModel):
        """
        Data describing a Person derived from an identity token provided
        by an authentication provider such as Auth0. Used during signup and login.
        """

        email: str
        name: str
        token: str



from radicli import Arg
from wasabi import msg

from ... import ty
from ...auth import get_current_auth
from ...cli import cli
from ...config import SavedSettings
from ...messages import Messages
from ...prodigy_teams_broker_sdk import models as broker_models
from ...ui import print_as_json
from ...util import resolve_remote_path


@cli.subcommand(
    "files",
    "ls",
    remote=Arg(help=Messages.remote_path),
    recurse=Arg("--recurse", "-r", help=Messages.recurse_list),
    as_json=Arg("--json", help=Messages.as_json),
    cluster_host=Arg("--cluster-host", help=Messages.cluster_host),
)
def ls(
    remote: str,
    recurse: bool = False,
    as_json: bool = False,
    cluster_host: ty.Optional[str] = None,
) -> broker_models.PathList:
    """List the files under `remote`"""
    broker_host = SavedSettings.get_and_save("broker_host", cluster_host)
    auth = get_current_auth()
    path = resolve_remote_path(auth.client, remote, broker_host)
    body = broker_models.Listing(path=path, recurse=recurse, include_stats=False)
    files = auth.broker_client.files.list_dir(body)
    if as_json:
        print_as_json(files.dict())
    else:
        msg.info(remote)
        for file_path in files.paths:
            print(file_path)
    return files



import inspect
import json
from pathlib import Path

import pytest

from prodigy_teams.__main__ import FILE
from prodigy_teams.about import __version__
from prodigy_teams.cli import document_cli
from prodigy_teams.main import cli

ROOT_PATH = Path(__file__).parent.parent
README_PATH = ROOT_PATH / "README.md"


def test_readme_up_to_date():
    with README_PATH.open("r", encoding="utf8") as f:
        existing = f.read()
    docs = document_cli(cli, root=ROOT_PATH)
    assert docs == existing, "README.md out of date, re-run pdcli dev document-clis"


def test_static_up_to_date():
    """Test that the static JSON matches the current live CLI."""
    err = "static CLI doesn't match live CLI, re-run pdcli dev update-ptc"
    with FILE.open("r", encoding="utf8") as f:
        existing = f.read()
    current = json.dumps(cli.to_static_json())
    assert existing == current, err


def test_ptc_cli_help_texts(command):
    """Test that all commands provide docstrings and argument help texts."""
    assert command.description, f"no docstring for {command.display_name}"
    for arg in command.args:
        if arg.id == cli.extra_key:
            continue
        assert arg.arg.help, f"no help text for {command.display_name} -> {arg.id}"


def test_pt_cli_all_annotated(command_no_placeholder):
    """Assert that all function arguments are annotated."""
    command = command_no_placeholder
    sig = inspect.signature(command.func)
    args = [a.id for a in command.args]
    params = [p for p in sig.parameters]
    assert sorted(args) == sorted(params), f"arg mismatch in {command.display_name}"


def test_ptc_cli_default_help_info(capsys):
    with pytest.raises(SystemExit):
        cli.run(["ptc"])
    captured = capsys.readouterr()
    expected_output_slice = """Prodigy Teams Command Line Interface."""
    assert expected_output_slice in captured.out


def test_version(capsys):
    with pytest.raises(SystemExit):
        cli.run(["ptc", "--version"])
    captured = capsys.readouterr()
    assert captured.out.strip() == __version__



from ..models import OrgCreating, OrgDeleting, OrgReading, OrgReturning, OrgUpdating
from .base import ModelClient


class Org(
    ModelClient[
        OrgCreating, OrgReading, OrgUpdating, OrgDeleting, OrgReturning, OrgReturning
    ]
):
    Creating = OrgCreating
    Reading = OrgReading
    Updating = OrgUpdating
    Deleting = OrgDeleting
    Returning = OrgReturning



from radicli import Arg
from wasabi import msg

from .. import ty
from ..cli import cli
from ..config import SavedSettings, config_dir
from ..messages import Messages
from ..query import resolve_action_id, resolve_project_id, resolve_task_id
from ..util import URL


@cli.subcommand("config", "reset")
def reset() -> None:
    """Reset all caching and configuration."""
    queue = list(config_dir().iterdir())
    for subpath in queue:
        if subpath.is_dir():
            queue.extend(list(subpath.iterdir()))
        else:
            msg.info(Messages.T021.format(subpath=subpath))
            subpath.unlink()


@cli.subcommand(
    "config",
    "project",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="project")),
)
def project(name_or_id: ty.StrOrUUID) -> ty.UUID:
    """Set the default project."""
    project_id = resolve_project_id(name_or_id)
    SavedSettings.get_and_save("project", project_id)
    msg.good(Messages.T019.format(noun="project", name=project_id))
    return project_id


@cli.subcommand(
    "config",
    "task",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="task")),
    project_id=Arg(help=Messages.project_id.format(noun="task")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="task")),
)
def task(
    name_or_id: ty.StrOrUUID,
    cluster_id: ty.Optional[ty.UUID] = None,
    project_id: ty.Optional[ty.UUID] = None,
) -> ty.UUID:
    """Set the default task."""
    task_id = resolve_task_id(name_or_id, project_id=project_id, broker_id=cluster_id)
    SavedSettings.get_and_save("task", task_id)
    msg.good(Messages.T019.format(noun="task", name=task_id))
    return task_id


@cli.subcommand(
    "config",
    "action",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="action")),
    project_id=Arg(help=Messages.project_id.format(noun="action")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="action")),
)
def action(
    name_or_id: ty.StrOrUUID,
    cluster_id: ty.Optional[ty.UUID] = None,
    project_id: ty.Optional[ty.UUID] = None,
) -> ty.UUID:
    """Set the default action."""
    action_id = resolve_action_id(
        name_or_id, project_id=project_id, broker_id=cluster_id
    )
    SavedSettings.get_and_save("action", action_id)
    msg.good(Messages.T019.format(noun="action", name=action_id))
    return action_id


@cli.subcommand(
    "config",
    "set-cluster-host",
    host=Arg(help=Messages.cluster_host_config),
)
def set_broker_host(host: str) -> None:
    """Set the broker cluster host."""
    host_url = URL.parse(host)
    SavedSettings.get_and_save("broker_host", str(host_url))
    msg.good(Messages.T019.format(noun="cluster host", name=host))


@cli.subcommand(
    "config",
    "set-pam-host",
    host=Arg(help=Messages.pam_host_config),
)
def set_pam_host(host: str) -> None:
    """Set the PAM host."""
    host_url = URL.parse(host)
    SavedSettings.get_and_save("pam_host", str(host_url))
    msg.good(Messages.T019.format(noun="PAM host", name=host))



from ..models import (
    InvitationCreating,
    InvitationDeleting,
    InvitationReading,
    InvitationReturning,
    InvitationUpdating,
)
from .base import ModelClient


class Invitation(
    ModelClient[
        InvitationCreating,
        InvitationReading,
        InvitationUpdating,
        InvitationDeleting,
        InvitationReturning,
        InvitationReturning,
    ]
):
    Creating = InvitationCreating
    Reading = InvitationReading
    Updating = InvitationUpdating
    Deleting = InvitationDeleting
    Returning = InvitationReturning



import os
import re
import sys
from dataclasses import dataclass
from urllib.parse import urlparse
from uuid import UUID

from . import ty
from .messages import Messages
from .prodigy_teams_pam_sdk import Client
from .prodigy_teams_pam_sdk.models import BrokerPathReading

APP_NAME = "prodigy-teams"
CYGWIN = sys.platform.startswith("cygwin")
MSYS2 = sys.platform.startswith("win") and ("GCC" in sys.version)
# Determine local App Engine environment, per Google's own suggestion
APP_ENGINE = "APPENGINE_RUNTIME" in os.environ and "Development/" in os.environ.get(
    "SERVER_SOFTWARE", ""
)
WIN = sys.platform.startswith("win") and not APP_ENGINE and not MSYS2


@dataclass(frozen=True)
class URL:
    # e.g. https://pam.explosion.rocks:8080/path
    url: str
    # e.g. pam.explosion.rocks:8080
    netloc: str
    # e.g. 8080, defaults to 80 for http and 443 for https
    port: int
    # e.g. https
    scheme: str

    @classmethod
    def parse(cls, host_or_url: str, scheme="https") -> "URL":
        host_or_url = host_or_url.strip().strip("/")
        if not host_or_url.startswith("http"):
            host_or_url = f"{scheme}://{host_or_url}"
        parse = urlparse(host_or_url)
        netloc = parse.netloc
        url = parse.geturl()
        scheme = parse.scheme
        port = (
            parse.port
            if parse.port is not None
            else cls._default_port_from_scheme(scheme)
        )
        return cls(url=url, netloc=netloc, port=port, scheme=scheme)

    @classmethod
    def _default_port_from_scheme(cls, scheme: str) -> int:
        if scheme == "https":
            return 443
        elif scheme == "http":
            return 80
        else:
            raise ValueError(f"unsupported url scheme: {scheme}")

    def __format__(self, __format_spec: str) -> str:
        return self.url

    def __str__(self) -> str:
        return self.url

    def __truediv__(self, sub_path: str) -> "URL":
        return URL(
            url=f"{self.url}/{sub_path.strip('/')}",
            netloc=self.netloc,
            port=self.port,
            scheme=self.scheme,
        )


# Source: https://github.com/pallets/click/blob/cba52fa76135af2edf46c154203b47106f898eb3/src/click/utils.py#L408
def get_app_dir(app_name: str, roaming: bool = True, force_posix: bool = False) -> str:
    """
    Returns the config folder for the application.  The default behavior
    is to return whatever is most appropriate for the operating system.
    """

    def _posixify(name: str) -> str:
        return "-".join(name.split()).lower()

    if WIN:
        key = "APPDATA" if roaming else "LOCALAPPDATA"
        folder = os.environ.get(key)
        if folder is None:
            folder = os.path.expanduser("~")
        return os.path.join(folder, app_name)
    if force_posix:
        return os.path.join(os.path.expanduser(f"~/.{_posixify(app_name)}"))
    if sys.platform == "darwin":
        return os.path.join(
            os.path.expanduser("~/Library/Application Support"), app_name
        )
    return os.path.join(
        os.environ.get("XDG_CONFIG_HOME", os.path.expanduser("~/.config")),
        _posixify(app_name),
    )


def is_local(path_str: str) -> bool:
    if path_str.startswith("{"):
        return False
    elif path_str.startswith("s3://"):
        return False
    elif path_str.startswith("gs://"):
        return False
    else:
        return True


def resolve_remote_path(client: Client, remote: str, default_broker: str) -> str:
    path_parts = parse_remote_path(remote)
    if path_parts[1] is None:
        return path_parts[2]
    broker_name, path_name, subpath = path_parts
    broker_name = broker_name or default_broker
    query = BrokerPathReading(
        broker_id=_resolve_broker_ref(client, broker_name),
        name=path_name,
        id=None,
        path=None,
    )
    result = client.broker_path.read(query)
    return os.path.join(result.path, subpath)


# Place this close to the function for convenience.
# TODO: This will have trouble with escaped curlies right?
_PATH_RE = re.compile(r"^(?:\{(\w+)\}:)?(?:\{(\w+)\}/?)?([^{}]*)$")


def parse_remote_path(path: str) -> ty.Tuple[ty.Optional[str], ty.Optional[str], str]:
    match_obj = _PATH_RE.match(path)
    if match_obj is None:
        raise ValueError(Messages.E128.format(path=path))
    groups = match_obj.groups()
    if len(groups) != 3:
        raise ValueError(Messages.E128.format(path=path))
    return groups


def _resolve_broker_ref(client: Client, name_or_id: str) -> ty.UUID:
    """
    Resolve reference to a task by name or ID. If an ID is given, return it.
    If it's a name, look up the ID.
    """
    try:
        return UUID(name_or_id)
    except ValueError:
        pass
    task = client.broker.read(name=name_or_id)
    return task.id



from ..models import (
    GroupAuthorizing,
    GroupCreating,
    GroupDeleting,
    GroupPopulating,
    GroupReading,
    GroupReturning,
    GroupUpdating,
)
from .base import ModelClient


class Group(
    ModelClient[
        GroupCreating,
        GroupReading,
        GroupUpdating,
        GroupDeleting,
        GroupReturning,
        GroupReturning,
    ]
):
    Creating = GroupCreating
    Reading = GroupReading
    Updating = GroupUpdating
    Deleting = GroupDeleting
    Returning = GroupReturning
    Populating = GroupPopulating
    Authorizing = GroupAuthorizing



from .. import ty
from ..models import Dataset, DatasetCreating, DatasetQueryResponse
from .base import BaseClient


class Data(BaseClient):
    def all(
        self,
        session: ty.Optional[bool] = None,
        *,
        page: ty.Optional[int] = None,
        size: ty.Optional[int] = None,
    ) -> ty.Page[str]:
        res = self.request(
            "GET",
            endpoint="all",
            data={"session": session},
            page=page,
            size=size,
            return_model=ty.Page[str],
        )
        return ty.cast(ty.Page[str], res)

    async def all_async(
        self,
        session: ty.Optional[bool] = None,
        *,
        page: ty.Optional[int] = None,
        size: ty.Optional[int] = None,
    ) -> ty.Page[str]:
        res = await self.request_async(
            "GET",
            endpoint="all",
            data={"session": session},
            page=page,
            size=size,
            return_model=ty.Page[str],
        )
        return ty.cast(ty.Page[str], res)

    def create(self, body: DatasetCreating) -> Dataset:
        res = self.request("POST", endpoint="create", data=body, return_model=Dataset)
        return ty.cast(Dataset, res)

    async def create_async(self, body: DatasetCreating) -> Dataset:
        res = await self.request_async(
            "POST", endpoint="create", data=body, return_model=Dataset
        )
        return ty.cast(Dataset, res)

    def read_examples(
        self,
        name: str,
        *,
        page: ty.Optional[int] = None,
        size: ty.Optional[int] = None,
    ) -> DatasetQueryResponse:
        params: dict = {
            "name": name,
        }
        if page is not None:
            params["page"] = page
        if size is not None:
            params["size"] = size

        res = self.request(
            "GET",
            endpoint="read-examples",
            params=params,
            return_model=DatasetQueryResponse,
        )
        return ty.cast(DatasetQueryResponse, res)

    async def read_examples_async(
        self,
        name: str,
        *,
        page: ty.Optional[int] = None,
        size: ty.Optional[int] = None,
    ) -> DatasetQueryResponse:
        params: dict = {
            "name": name,
        }
        if page is not None:
            params["page"] = page
        if size is not None:
            params["size"] = size

        res = await self.request_async(
            "GET",
            endpoint="read-examples",
            params=params,
            return_model=DatasetQueryResponse,
        )
        return ty.cast(DatasetQueryResponse, res)



from ..models import (
    ActionCreating,
    ActionDeleting,
    ActionReading,
    ActionReturning,
    ActionUpdating,
)
from .base import ModelClient


class Action(
    ModelClient[
        ActionCreating,
        ActionReading,
        ActionUpdating,
        ActionDeleting,
        ActionReturning,
        ActionReturning,
    ]
):
    Creating = ActionCreating
    Reading = ActionReading
    Updating = ActionUpdating
    Deleting = ActionDeleting
    Returning = ActionReturning



import json
import time

import httpx
from radicli import Arg
from wasabi import msg

from .. import ty
from ..auth import get_current_auth
from ..cli import cli
from ..errors import (
    ConnectError,
    ConnectTimeout,
    HTTPStatusError,
    InvalidURL,
    RequestError,
)
from ..messages import Messages
from ..prodigy_teams_broker_sdk.models import CheckProgressRequest, CheckStartRequest
from ..prodigy_teams_pam_sdk import models as prodigy_teams_pam_sdk_models
from ..prodigy_teams_pam_sdk.models import BrokerUpdating
from ..query import resolve_broker, resolve_broker_id, resolve_package, resolve_recipe
from ..ui import print_info_table, print_table_with_select
from ..util import URL

# Broker has wrong name on infra
# TODO: Basic crud stuff here for clusters


def check_broker_status(
    broker: prodigy_teams_pam_sdk_models.BrokerReturning,
) -> ty.BrokerStatusCheck:
    url = URL.parse(broker.address) / "api/v1/status"
    # Before querying the broker, verify that it has finished registration
    # TODO: also track when it is killed in PAM
    if broker.state == "creating":
        return ty.BrokerStatusCheck.CREATING
    try:
        response = httpx.get(str(url))
        response.raise_for_status()
        payload = response.json()
        # Happy case: successfully connect to the broker and check "cluster" key
        if payload["cluster"] == "Ready":
            return ty.BrokerStatusCheck.RUNNING
        else:
            return ty.BrokerStatusCheck.ISSUES
    except InvalidURL:
        return ty.BrokerStatusCheck.INVALID_ADDRESS
    except (ConnectTimeout, ConnectError):
        # This could mean the broker is down, but could also mean we have no network access
        return ty.BrokerStatusCheck.NOT_FOUND
    except RequestError as e:
        msg.warn(Messages.E041.format(name=broker.name), str(e))
        return ty.BrokerStatusCheck.REQUEST_ERROR
    except HTTPStatusError as e:
        # We shouldn't really log here -- better to refactor this into the commands
        msg.warn(Messages.E041.format(name=broker.name), str(e.response.status_code))
        return ty.BrokerStatusCheck.RESPONSE_ERROR


class BrokerReturning(prodigy_teams_pam_sdk_models.BrokerReturning):
    """
    Extends the data returned by the API with computed properties.
    This is a workaround for the models being auto-generated.
    """

    @property
    def status(self) -> str:
        status = check_broker_status(self)
        return str(status)

    @classmethod
    def get_properties(cls) -> ty.List[str]:
        return [
            prop for prop in cls.__dict__ if isinstance(cls.__dict__[prop], property)
        ]

    def _iter(
        self,
        to_dict: bool = False,
        by_alias: bool = False,
        include=None,
        exclude=None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        exclude_properties: bool = False,
    ) -> ty.Iterator[ty.Tuple[str, ty.Any]]:
        yield from super()._iter(
            to_dict=to_dict,
            by_alias=by_alias,
            include=include,  # type: ignore
            exclude=exclude,  # type: ignore
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        if not exclude_properties:
            props = self.get_properties()
            if include:
                props = [prop for prop in props if prop in include]
            if exclude:
                props = [prop for prop in props if prop not in exclude]
            for key in props:
                yield key, getattr(self, key)


@cli.subcommand(
    "clusters",
    "list",
    # fmt: off
    select=Arg("--select", help=Messages.select.format(opts=list(BrokerReturning.__fields__))),
    # fmt: on
)
def list(
    select: ty.List[str] = ["id", "name", "status", "address"]
) -> ty.List[BrokerReturning]:
    """List resources on the cluster"""
    client = get_current_auth().client
    # TODO: this sequentially checks broker status which could be slow
    res = [
        BrokerReturning(**x.dict()) for x in client.broker.all(page=1, size=100).items
    ]
    print_table_with_select(res, select=select)
    return res


@cli.subcommand(
    "clusters",
    "info",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="cluster")),
    as_json=Arg("--json", help=Messages.as_json),
)
def info(name_or_id: ty.StrOrUUID, as_json: bool = False) -> BrokerReturning:
    """Get detailed info for a cluster"""
    cluster = BrokerReturning(**resolve_broker(name_or_id).dict())
    select = [*BrokerReturning.__fields__.keys(), *BrokerReturning.get_properties()]
    print_info_table(cluster, select=select, as_json=as_json)
    return cluster


@cli.subcommand(
    "clusters",
    "update",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="cluster")),
    name=Arg("--new-name", help=Messages.new_name.format(noun="cluster")),
    address=Arg("--address", help=Messages.new_address.format(noun="cluster")),
    as_json=Arg("--json", help=Messages.as_json),
)
def update(
    name_or_id: ty.StrOrUUID,
    name: ty.Optional[str] = None,
    address: ty.Optional[str] = None,
    as_json: bool = False,
) -> ty.UUID:
    """Update the cluster info"""
    cluster = resolve_broker(name_or_id)
    # TODO: The BrokerUpdating model is currently wrong on pam, making these
    # fields required. Change the query when this is fixed.
    body = BrokerUpdating(
        id=cluster.id, name=name or cluster.name, address=address or cluster.address
    )
    auth = get_current_auth()
    res = auth.client.broker.update(body)
    print_info_table(res, as_json=as_json)
    return cluster.id


@cli.subcommand(
    "clusters",
    "delete",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="cluster")),
)
def delete(name_or_id: ty.StrOrUUID) -> ty.UUID:
    """
    Delete a cluster from PAM. This only removes PAM's record
    of it. The cluster itself will continue to exist - you need
    to shut it down separately.
    """
    auth = get_current_auth()
    client = auth.client
    cluster_id = resolve_broker_id(name_or_id)
    client.broker.delete(id=cluster_id)
    msg.good(Messages.T003.format(noun="cluster", name=cluster_id))
    return cluster_id


@cli.subcommand(
    "clusters",
    "check",
    s3_bucket=Arg("--s3-bucket", help=Messages.check_s3_bucket),
    nfs_path=Arg("--nfs-path", help=Messages.check_nfs_path),
    recipe_name_or_id=Arg("--recipe", help=Messages.check_recipe),
    recipe_args=Arg("--recipe-args", help=Messages.check_recipe_args),
)
def check(
    s3_bucket: ty.Optional[str] = None,
    nfs_path: ty.Optional[str] = None,
    recipe_name_or_id: ty.Optional[ty.StrOrUUID] = None,
    recipe_args: ty.Optional[str] = None,
) -> None:
    """Check the cluster deployment went well"""
    auth = get_current_auth()
    if recipe_args:
        recipe_args_dict = json.loads(recipe_args)
        assert isinstance(recipe_args_dict, dict)
    else:
        recipe_args_dict = {}
    if recipe_name_or_id:
        recipe = resolve_recipe(recipe_name_or_id, broker_id=None)
        package = resolve_package(recipe.package_id, broker_id=None)
        recipe_name = recipe.name
        package_environment = package.environment
    else:
        recipe_name = None
        package_environment = None
    body = CheckStartRequest(
        s3_bucket=s3_bucket,
        nfs_path=nfs_path,
        package_environment=package_environment,
        recipe_name=recipe_name,
        recipe_args=recipe_args_dict,
    )
    start_response = auth.broker_client.check.start(body)
    while True:
        body = CheckProgressRequest(id=start_response.id)
        progress_response = auth.broker_client.check.progress(body)
        if progress_response.status == "done":
            assert progress_response.report
            msg.table(
                [x for x in progress_response.report.items()],  # type: ignore
                header=["Check Name", "Result"],
                divider=True,
            )
            break
        time.sleep(1)



import logging
import os
import time
from pathlib import Path

import httpx
import pydantic
from pydantic import BaseModel

from . import ty
from .config import SavedSettings, config_dir
from .errors import (
    CLIError,
    HTTPError,
    ProdigyTeamsError,
    ProdigyTeamsIDTokenError,
    ProdigyTeamsParseSecretsError,
)
from .messages import Messages
from .prodigy_teams_broker_sdk import (
    AccessTokenCredential as BrokerAccessTokenCredential,
)
from .prodigy_teams_broker_sdk import Client as BrokerClient
from .prodigy_teams_pam_sdk import AccessTokenCredential, Client
from .prodigy_teams_pam_sdk.models import BrokerReading
from .ui import print_login_info
from .util import URL

logger = logging.getLogger(__name__)


AUTH0_SCOPE = "openid profile email"
AUTH0_GRANT_TYPE = "urn:ietf:params:oauth:grant-type:device_code"
AUTH_WAIT_TIME_SECONDS = 60 * 5


class Secrets(ty.Protocol):
    """
    Defines the interface for secrets storage. During normal operation,
    the secrets are stored in a file on disk. During testing, a mock
    implementation can be used to prevent breaking configuration.
    """

    id_token: ty.Optional[str]
    api_token: ty.Optional[AccessTokenCredential]
    broker_tokens: ty.Dict[str, BrokerAccessTokenCredential]

    @classmethod
    def load(cls) -> "Secrets":
        ...

    @classmethod
    def clean(cls) -> None:
        ...

    def save(self) -> None:
        ...


class FileSecrets(BaseModel):
    id_token: ty.Optional[str] = None
    api_token: ty.Optional[AccessTokenCredential] = None
    broker_tokens: ty.Dict[str, BrokerAccessTokenCredential] = {}

    @classmethod
    def _path(cls) -> Path:
        return config_dir() / "secrets.json"

    @classmethod
    def load(cls) -> "Secrets":
        try:
            return cls.parse_file(cls._path())
        except pydantic.error_wrappers.ValidationError as e:
            raise ProdigyTeamsParseSecretsError(
                secrets_file=cls._path(), error=e
            ) from e
        except FileNotFoundError:
            return cls()

    @classmethod
    def clean(cls) -> None:
        cls._path().unlink(missing_ok=True)

    def save(self) -> None:
        path = self._path()
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open(mode="w", encoding="utf-8") as f:
            try:
                os.fchmod(f.fileno(), 0o600)
            except AttributeError:
                # no chmod on Windows
                pass
            f.write(self.json())


class CLIAuth(ty.Protocol):
    @property
    def client(self) -> Client:
        ...

    @property
    def broker_client(self) -> BrokerClient:
        ...

    @property
    def broker_host(self) -> str:
        ...

    @property
    def broker_url(self) -> URL:
        ...

    @property
    def broker_id(self) -> ty.UUID:
        ...

    @property
    def org_id(self) -> ty.UUID:
        ...

    @property
    def pam_host(self) -> str:
        ...

    @property
    def pam_url(self) -> URL:
        ...

    @property
    def secrets(self) -> Secrets:
        ...

    def retry_id(self, func) -> ty.Callable[..., ty.Any]:
        ...

    def retry_api(self, func) -> ty.Callable[..., ty.Any]:
        ...

    def retry(self, func) -> ty.Callable[..., ty.Any]:
        ...

    def _ensure_readable_secrets(self) -> Secrets:
        ...

    def _ensure_broker_host(self) -> None:
        ...

    def get_id_token(self, force_refresh: bool = False) -> str:
        ...

    def get_api_token(self, force_refresh: bool = False) -> AccessTokenCredential:
        ...

    def get_broker_token(
        self, force_refresh: bool = False
    ) -> BrokerAccessTokenCredential:
        ...


class CLIAuthImpl:
    _pam_url: URL
    _broker_url: ty.Optional[URL]
    _broker_id: ty.Optional[ty.UUID]
    _org_id: ty.Optional[ty.UUID]

    def __init__(self, *, pam_host: str, broker_host: ty.Optional[str]) -> None:
        self._secrets: ty.Optional[Secrets] = None
        self._broker_url = URL.parse(broker_host) if broker_host is not None else None
        self._pam_url = URL.parse(pam_host)
        self._broker_id = None
        self._org_id = None

    @property
    def client(self) -> Client:
        return Client(base_url=self.pam_url.url, token=self.get_api_token())

    @property
    def broker_client(self) -> BrokerClient:
        self._ensure_broker_host()
        return BrokerClient(base_url=self.broker_url.url, token=self.get_broker_token())

    @property
    def broker_host(self) -> str:
        return self.broker_url.netloc

    @property
    def broker_url(self) -> URL:
        assert self._broker_url is not None
        return self._broker_url

    @property
    def broker_id(self) -> ty.UUID:
        if self._broker_id is None:
            if self._broker_url is None:
                raise CLIError(Messages.E035)
            # The models from pam_sdk need to be generated using the default keyword
            # when using pydantic.Field. Pyright needs this to work well
            # Ref: https://pydantic-docs.helpmanual.io/visual_studio_code/#adding-a-default-with-field
            body = BrokerReading(address=str(self.broker_url))  # pyright: ignore
            broker = self.client.broker.read(body)
            self._broker_id = broker.id
        return self._broker_id

    @property
    def org_id(self) -> ty.UUID:
        if self._org_id is None:
            broker = self.client.broker.read(id=self.broker_id)
            self._org_id = broker.org_id
        return self._org_id

    @property
    def pam_host(self) -> str:
        return self.pam_url.netloc

    @property
    def pam_url(self) -> URL:
        # TODO: better error message
        assert self._pam_url is not None
        return self._pam_url

    @property
    def secrets(self) -> Secrets:
        # Secrets are lazy loaded to prevent `ptc login` from crashing due to
        # a broken secrets file.
        if self._secrets is None:
            self._secrets = FileSecrets.load()
        return self._secrets

    def retry_id(self, func) -> ty.Callable[..., ty.Any]:
        def wrapper(*args, **kwargs) -> ty.Any:
            try:
                return func(*args, **kwargs)
            except ProdigyTeamsIDTokenError:
                self.get_id_token(force_refresh=True)
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    logger.exception(e)
                    raise CLIError(Messages.E104, e)

        return wrapper

    def retry_api(self, func) -> ty.Callable[..., ty.Any]:
        def wrapper(*args, **kwargs) -> ty.Any:
            try:
                return func(*args, **kwargs)
            except HTTPError:
                self.get_api_token(force_refresh=True)
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    logger.exception(e)
                    raise CLIError(Messages.E105, e)

        return wrapper

    def retry(self, func) -> ty.Callable[..., ty.Any]:
        @self.retry_api
        @self.retry_id
        def wrapper(*args, **kwargs) -> ty.Any:
            return func(*args, **kwargs)

        return wrapper

    def _ensure_readable_secrets(self) -> Secrets:
        """Try to load the secrets file, cleaning the file if there is an error"""
        self._secrets = None
        try:
            self.secrets
        except ProdigyTeamsParseSecretsError:
            FileSecrets.clean()
        return self.secrets

    def _ensure_broker_host(self) -> None:
        if self._broker_url is None:
            brokers = self.client.broker.all(size=2)
            if brokers.total == 0:
                raise CLIError(Messages.E106)
            elif brokers.total == 1:
                self._broker_url = URL.parse(brokers.items[0].address)
                SavedSettings.get_and_save("broker_host", self.broker_host)
            else:
                raise CLIError(Messages.E107)

    @staticmethod
    def _check_token_expired(
        token: ty.Union[AccessTokenCredential, BrokerAccessTokenCredential, str]
    ) -> bool:
        # TODO: this is reusing the pam_sdk implementation to reduce duplication
        # in the tricky token management logic. Needs to be cleaned up.
        if isinstance(token, str):
            return AccessTokenCredential(access_token=token).expired
        return token.expired

    def get_id_token(self, force_refresh: bool = False) -> str:
        if (
            self.secrets.id_token is None
            or self._check_token_expired(self.secrets.id_token)
            or force_refresh
        ):
            id_token = self._authenticate_device_and_get_id_token()
            self.secrets.id_token = id_token
            self.secrets.save()
            return id_token
        else:
            return self.secrets.id_token

    def get_api_token(self, force_refresh: bool = False) -> AccessTokenCredential:
        if (
            self.secrets.api_token is None
            or self.secrets.api_token.expired
            or force_refresh
        ):
            id_token = self.get_id_token()
            api_response = httpx.post(
                f"{self.pam_url}/api/v1/login",
                headers={"authorization": f"Bearer {id_token}"},
            )
            if api_response.status_code != 200:
                breakpoint()
                raise ProdigyTeamsIDTokenError()
            data = api_response.json()
            token = AccessTokenCredential(access_token=data["access_token"])
            self.secrets.api_token = token
            self.secrets.save()
            return token
        else:
            return self.secrets.api_token

    def get_broker_token(
        self, force_refresh: bool = False
    ) -> BrokerAccessTokenCredential:
        self._ensure_broker_host()
        if (
            self.broker_host not in self.secrets.broker_tokens
            or self._check_token_expired(self.secrets.broker_tokens[self.broker_host])
            or force_refresh
        ):
            api_token = self.get_api_token().access_token
            res = httpx.post(
                f"{self.broker_url}/api/v1/token",
                headers={"authorization": f"Bearer {api_token}"},
            )
            if res.status_code != 200:
                err = Messages.E034.format(
                    code=res.status_code, reason=res.reason_phrase, url=res.url
                )
                if res.status_code == 401:
                    raise ProdigyTeamsIDTokenError(err)
                raise ProdigyTeamsError(err)
            data = res.json()
            token = BrokerAccessTokenCredential(access_token=data["access_token"])
            self.secrets.broker_tokens[self.broker_host] = token
            self.secrets.save()
            return token
        else:
            return self.secrets.broker_tokens[self.broker_host]

    def _authenticate_device_and_get_id_token(self) -> str:
        try:
            config_response = httpx.get(f"{self.pam_url}/api/v1/cli.json")
            config_response.raise_for_status()
        except Exception as exc:
            err = Messages.E033.format(url=self.pam_url)
            logger.exception(err)
            raise ProdigyTeamsError(err) from exc
        config = config_response.json()
        post_data = {
            "client_id": config["dclient_id"],
            "scope": AUTH0_SCOPE,
            "audience": config["daudience"],
        }
        response = httpx.post(config["dcode_url"], data=post_data)
        if response.status_code != 200:
            error_resp = response.json()
            raise CLIError(Messages.E101, error_resp)
        data = response.json()
        device_code = data.get("device_code")
        user_code = data.get("user_code")
        uri = data.get("verification_uri")
        data.get("expires_in")
        interval = data.get("interval", 5)
        uri_complete = data.get("verification_uri_complete")
        if not (device_code and user_code and uri and interval and uri_complete):
            raise CLIError(Messages.E032, data)
        print_login_info(uri_complete, uri, user_code)
        last_check = 0
        for _ in range(AUTH_WAIT_TIME_SECONDS):
            if time.time() > (last_check + int(interval)):
                token_response = httpx.post(
                    config["dtoken_url"],
                    data={
                        "grant_type": AUTH0_GRANT_TYPE,
                        "device_code": device_code,
                        "client_id": config["dclient_id"],
                    },
                )
                last_check = time.time()
                token_data = token_response.json()
                if token_response.status_code == 200:
                    return token_data.get("id_token")
            time.sleep(1)
        raise ProdigyTeamsIDTokenError(Messages.E031)


AUTH = None


def get_current_auth() -> "CLIAuth":
    """Fetches the current CLI Auth."""
    global AUTH
    if AUTH is None:
        settings = SavedSettings.load()
        if settings.pam_host is None:
            raise CLIError(Messages.E037)
        AUTH = CLIAuthImpl(pam_host=settings.pam_host, broker_host=settings.broker_host)
    return AUTH



from ..models import (
    UserCreating,
    UserDeleting,
    UserReading,
    UserReturning,
    UserUpdating,
)
from .base import ModelClient


class User(
    ModelClient[
        UserCreating,
        UserReading,
        UserUpdating,
        UserDeleting,
        UserReturning,
        UserReturning,
    ]
):
    Creating = UserCreating
    Reading = UserReading
    Updating = UserUpdating
    Deleting = UserDeleting
    Returning = UserReturning



from .. import ty
from ..models import (
    CheckProgressRequest,
    CheckProgressResponse,
    CheckStartRequest,
    CheckStartResponse,
)
from .base import BaseClient


class Check(BaseClient):
    def start(self, body: CheckStartRequest) -> CheckStartResponse:
        res = self.request(
            "POST", endpoint="start", data=body, return_model=CheckStartResponse
        )
        return ty.cast(CheckStartResponse, res)

    async def start_async(self, body: CheckStartRequest) -> CheckStartResponse:
        res = await self.request_async(
            "POST", endpoint="start", data=body, return_model=CheckStartResponse
        )
        return ty.cast(CheckStartResponse, res)

    def progress(self, body: CheckProgressRequest) -> CheckProgressResponse:
        res = self.request(
            "POST", endpoint="progress", data=body, return_model=CheckProgressResponse
        )
        return ty.cast(CheckProgressResponse, res)

    async def progress_async(self, body: CheckProgressRequest) -> CheckProgressResponse:
        res = await self.request_async(
            "POST", endpoint="progress", data=body, return_model=CheckProgressResponse
        )
        return ty.cast(CheckProgressResponse, res)



import os
import os.path
from pathlib import Path

from radicli import Arg
from wasabi import msg

from ... import ty
from ...auth import get_current_auth
from ...cli import cli
from ...config import SavedSettings
from ...errors import BrokerError, CLIError
from ...messages import Messages
from ...prodigy_teams_broker_sdk import Client as BrokerClient
from ...prodigy_teams_broker_sdk import models as broker_models
from ...ui import Progress, is_silent
from ...util import is_local, resolve_remote_path


class CopyCallable(ty.Protocol):
    def __call__(
        self,
        *,
        src: str,
        dest: str,
        overwrite: bool,
        recurse: bool,
        make_dirs: bool,
        broker_client: BrokerClient,
    ) -> ty.Optional[str]:
        ...


@cli.subcommand(
    "files",
    "cp",
    src=Arg(help=Messages.remote_local_path.format(noun="source")),
    dest=Arg(help=Messages.remote_local_path.format(noun="destination")),
    recurse=Arg("--recurse", "-r", help=Messages.recurse_copy),
    make_dirs=Arg("--make-dirs", help=Messages.make_dirs),
    overwrite=Arg("--overwrite", help=Messages.overwrite),
    cluster_host=Arg("--cluster-host", help=Messages.cluster_host),
)
def cp(
    src: str,
    dest: str,
    recurse: bool = False,
    make_dirs: bool = False,
    overwrite: bool = False,
    cluster_host: ty.Optional[str] = None,
) -> None:
    """Copy files to and from the cluster"""
    broker_host = SavedSettings.get_and_save("broker_host", cluster_host)
    auth = get_current_auth()
    is_src_local = is_local(src)
    is_dest_local = is_local(dest)

    def _cp(f: CopyCallable, *, src: str, dest: str) -> ty.Optional[str]:
        return f(
            src=src,
            dest=dest,
            overwrite=overwrite,
            recurse=recurse,
            make_dirs=make_dirs,
            broker_client=auth.broker_client,
        )

    if is_src_local and is_dest_local:  # both local
        raise CLIError(Messages.E015.format(verb="copy"), f"{src}, {dest}")
    elif is_src_local and not is_dest_local:  # from local to remote
        err = _cp(
            _upload,
            src=src,
            dest=resolve_remote_path(auth.client, dest, broker_host),
        )
    elif not is_src_local and is_dest_local:
        # From remote to local
        err = _cp(
            _download,
            src=resolve_remote_path(auth.client, src, broker_host),
            dest=dest,
        )
    else:
        # Both remote
        err = _cp(
            _copy_on_remote,
            src=resolve_remote_path(auth.client, src, broker_host),
            dest=resolve_remote_path(auth.client, dest, broker_host),
        )
    if err is not None:
        raise CLIError(err)
    elif not is_silent():
        msg.good(Messages.T011, f"{src} -> {dest}")


def _upload(
    *,
    src: str,
    dest: str,
    make_dirs: bool,
    overwrite: bool,
    recurse: bool,
    broker_client: BrokerClient,
) -> ty.Optional[str]:
    src_path = Path(src)
    if src_path.is_dir() and not recurse:
        raise CLIError(Messages.E016, src_path)
    if src_path.is_dir():
        paths = [path for path in src_path.glob("**/*") if path.is_file()]
    else:
        paths = [src_path]
        src_path = src_path.parent

    def get_dest(path):
        return os.path.join(dest, str(path.relative_to(src_path)))

    with Progress(paths) as paths:
        for path in paths:
            body = path.open("rb")
            try:
                broker_client.files.upload(
                    body, dest=get_dest(path), overwrite=overwrite, make_dirs=make_dirs
                )
            except BrokerError as e:
                raise CLIError(Messages.E017.format(verb="upload"), e)


def _download(
    *,
    src: str,
    dest: str,
    make_dirs: bool,
    overwrite: bool,
    recurse: bool,
    broker_client: BrokerClient,
) -> ty.Optional[str]:
    body = broker_models.Listing(path=src, recurse=recurse, include_stats=False)
    try:
        files = broker_client.files.list_dir(body)
    except BrokerError as e:
        raise CLIError(Messages.E018, e)
    if not recurse:
        if len(files.paths) > 1:
            raise CLIError(Messages.E019, src)

    get_dest: ty.Callable[[str], Path] = (
        (lambda _: Path(dest))
        if not recurse
        # Important: we need to do str(Path()) here to ensure a trailing slash is
        # included. Otherwise, we end up with x / '/foo', which becomes '/foo'
        # and NOT x/foo, which would be expected!
        else lambda path: Path(dest) / (path.removeprefix(str(Path(src))))
    )
    with Progress(files.paths) as paths:
        for path in paths:
            dest_ = get_dest(path)
            if not overwrite and dest_.exists():
                raise CLIError(Messages.E020, dest_)
            if make_dirs:
                dest_.parent.mkdir(parents=True, exist_ok=True)
            elif not dest_.parent.exists():
                raise CLIError(Messages.E021, dest_)
            body = broker_models.Downloading(target=path)
            try:
                content = broker_client.files.download(body)
            except BrokerError as e:
                raise CLIError(Messages.E017.format(verb="download"), e)
            with dest_.open("wb") as file_:
                file_.write(content.read())


def _copy_on_remote(
    *,
    src: str,
    dest: str,
    make_dirs: bool,
    overwrite: bool,
    recurse: bool,
    broker_client: BrokerClient,
) -> ty.Optional[str]:
    if recurse:
        body = broker_models.Copying(
            src=src, dest=dest, make_dirs=make_dirs, overwrite=overwrite
        )
        try:
            plan = broker_client.files.plan_directory_copy(body)
        except BrokerError as e:
            raise CLIError(Messages.E022, e)
        copies = plan.copy_
    else:
        copies = [
            broker_models.Copying(
                src=src, dest=dest, make_dirs=make_dirs, overwrite=overwrite
            )
        ]
    with Progress(copies) as copies:
        for file_copy_plan in copies:
            body = broker_models.Copying(
                src=file_copy_plan.src,
                dest=file_copy_plan.dest,
                make_dirs=make_dirs,
                overwrite=overwrite,
            )
            try:
                broker_client.files.copy(body)
            except BrokerError as e:
                raise CLIError(Messages.E017.format(verb="copy"), e)



from . import ty


class ProdigyTeamsError(Exception):
    def __init__(
        self,
        message: ty.Optional[str] = None,
        detail: ty.Optional[str] = None,
        status: ty.Optional[int] = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__()
        self.message = message
        self.detail = detail
        self.status = status
        self.args = args
        self.kwargs = kwargs

    def __str__(self) -> str:
        output = ""
        if self.message:
            output = self.message
        if self.detail:
            output += f"\n{self.detail}" if output else self.detail
        return output


class RecipeProcessingError(Exception):
    def __init__(self, title: str, text: ty.Optional[ty.Any] = None) -> None:
        self.title = title
        self.text = text
        self.message = self.title + (f"\n{self.text}" if self.text else "")
        super().__init__(self.message)


class BaseCRUDError(ProdigyTeamsError):
    """Base class that can be changed per project."""


class Invalid(BaseCRUDError):
    pass


class NotFound(BaseCRUDError):
    pass


class Exists(BaseCRUDError):
    pass


class Forbidden(BaseCRUDError):
    pass


class Unsupported(BaseCRUDError):
    pass


class ForbiddenCreate(Forbidden):
    pass


class ForbiddenRead(Forbidden):
    pass


class ForbiddenUpdate(Forbidden):
    pass


class ForbiddenDelete(Forbidden):
    pass


class AuthError(ProdigyTeamsError):
    def __init__(self, detail: str = "Auth Error"):
        self.status = 401
        self.detail = detail

    def __str__(self):
        return self.detail


class AssetNotFound(NotFound):
    pass


class AssetInvalid(Invalid):
    pass


class AssetExists(Exists):
    pass


class AssetForbiddenCreate(ForbiddenCreate):
    pass


class AssetForbiddenRead(ForbiddenRead):
    pass


class AssetForbiddenUpdate(ForbiddenUpdate):
    pass


class AssetForbiddenDelete(ForbiddenDelete):
    pass


class AssetUnsupported(Unsupported):
    pass


class EnvironmentNotFound(NotFound):
    pass


class EnvironmentInvalid(Invalid):
    pass


class EnvironmentExists(Exists):
    pass


class EnvironmentForbiddenCreate(ForbiddenCreate):
    pass


class EnvironmentForbiddenRead(ForbiddenRead):
    pass


class EnvironmentForbiddenUpdate(ForbiddenUpdate):
    pass


class EnvironmentForbiddenDelete(ForbiddenDelete):
    pass


class EnvironmentUnsupported(Unsupported):
    pass


class GroupNotFound(NotFound):
    pass


class GroupInvalid(Invalid):
    pass


class GroupExists(Exists):
    pass


class GroupForbiddenCreate(ForbiddenCreate):
    pass


class GroupForbiddenRead(ForbiddenRead):
    pass


class GroupForbiddenUpdate(ForbiddenUpdate):
    pass


class GroupForbiddenDelete(ForbiddenDelete):
    pass


class GroupUnsupported(Unsupported):
    pass


class BrokerRegistrationNotFound(NotFound):
    pass


class BrokerRegistrationInvalid(Invalid):
    pass


class BrokerRegistrationExists(Exists):
    pass


class BrokerRegistrationForbiddenCreate(ForbiddenCreate):
    pass


class BrokerRegistrationForbiddenRead(ForbiddenRead):
    pass


class BrokerRegistrationForbiddenUpdate(ForbiddenUpdate):
    pass


class BrokerRegistrationForbiddenDelete(ForbiddenDelete):
    pass


class BrokerRegistrationUnsupported(Unsupported):
    pass


class BrokerNotFound(NotFound):
    pass


class BrokerInvalid(Invalid):
    pass


class BrokerExists(Exists):
    pass


class BrokerForbiddenCreate(ForbiddenCreate):
    pass


class BrokerForbiddenRead(ForbiddenRead):
    pass


class BrokerForbiddenUpdate(ForbiddenUpdate):
    pass


class BrokerForbiddenDelete(ForbiddenDelete):
    pass


class BrokerUnsupported(Unsupported):
    pass


class BrokerPathNotFound(NotFound):
    pass


class BrokerPathInvalid(Invalid):
    pass


class BrokerPathExists(Exists):
    pass


class BrokerPathForbiddenCreate(ForbiddenCreate):
    pass


class BrokerPathForbiddenRead(ForbiddenRead):
    pass


class BrokerPathForbiddenUpdate(ForbiddenUpdate):
    pass


class BrokerPathForbiddenDelete(ForbiddenDelete):
    pass


class BrokerPathUnsupported(Unsupported):
    pass


class OrgNotFound(NotFound):
    pass


class OrgInvalid(Invalid):
    pass


class OrgExists(Exists):
    pass


class OrgForbiddenCreate(ForbiddenCreate):
    pass


class OrgForbiddenRead(ForbiddenRead):
    pass


class OrgForbiddenUpdate(ForbiddenUpdate):
    pass


class OrgForbiddenDelete(ForbiddenDelete):
    pass


class OrgUnsupported(Unsupported):
    pass


class PackageNotFound(NotFound):
    pass


class PackageInvalid(Invalid):
    pass


class PackageExists(Exists):
    pass


class PackageForbiddenCreate(ForbiddenCreate):
    pass


class PackageForbiddenRead(ForbiddenRead):
    pass


class PackageForbiddenUpdate(ForbiddenUpdate):
    pass


class PackageForbiddenDelete(ForbiddenDelete):
    pass


class PackageForbiddenDeleteForExistingPlans(PackageForbiddenDelete):
    pass


class PackageUnsupported(Unsupported):
    pass


class PersonNotFound(NotFound):
    pass


class PersonInvalid(Invalid):
    pass


class PersonExists(Exists):
    pass


class PersonForbiddenCreate(ForbiddenCreate):
    pass


class PersonForbiddenRead(ForbiddenRead):
    pass


class PersonForbiddenUpdate(ForbiddenUpdate):
    pass


class PersonForbiddenDelete(ForbiddenDelete):
    pass


class PersonUnsupported(Unsupported):
    pass


class ProjectNotFound(NotFound):
    pass


class ProjectInvalid(Invalid):
    pass


class ProjectExists(Exists):
    pass


class ProjectForbiddenCreate(ForbiddenCreate):
    pass


class ProjectForbiddenRead(ForbiddenRead):
    pass


class ProjectForbiddenUpdate(ForbiddenUpdate):
    pass


class ProjectForbiddenDelete(ForbiddenDelete):
    pass


class ProjectUnsupported(Unsupported):
    pass


class RecipeNotFound(NotFound):
    pass


class RecipeInvalid(Invalid):
    pass


class RecipeExists(Exists):
    pass


class RecipeForbiddenCreate(ForbiddenCreate):
    pass


class RecipeForbiddenRead(ForbiddenRead):
    pass


class RecipeForbiddenUpdate(ForbiddenUpdate):
    pass


class RecipeForbiddenDelete(ForbiddenDelete):
    pass


class RecipeUnsupported(Unsupported):
    pass


class Recipe_AmbiguousReference(Invalid):
    pass


class RequirementsGroupNotFound(NotFound):
    pass


class RequirementsGroupInvalid(Invalid):
    pass


class RequirementsGroupExists(Exists):
    pass


class RequirementsGroupForbiddenCreate(ForbiddenCreate):
    pass


class RequirementsGroupForbiddenRead(ForbiddenRead):
    pass


class RequirementsGroupForbiddenUpdate(ForbiddenUpdate):
    pass


class RequirementsGroupForbiddenDelete(ForbiddenDelete):
    pass


class RequirementsGroupUnsupported(Unsupported):
    pass


class InvitationNotFound(NotFound):
    pass


class InvitationInvalid(Invalid):
    pass


class InvitationExists(Exists):
    pass


class InvitationForbiddenCreate(ForbiddenCreate):
    pass


class InvitationForbiddenRead(ForbiddenRead):
    pass


class InvitationForbiddenUpdate(ForbiddenUpdate):
    pass


class InvitationForbiddenDelete(ForbiddenDelete):
    pass


class InvitationUnsupported(Unsupported):
    pass


class JobPlanNotFound(NotFound):
    pass


class JobPlanInvalid(Invalid):
    pass


class JobPlanExists(Exists):
    pass


class JobPlanForbiddenCreate(ForbiddenCreate):
    pass


class JobPlanForbiddenRead(ForbiddenRead):
    pass


class JobPlanForbiddenUpdate(ForbiddenUpdate):
    pass


class JobPlanForbiddenDelete(ForbiddenDelete):
    pass


class JobPlanUnsupported(Unsupported):
    pass


class JobArgNotFound(NotFound):
    pass


class JobArgInvalid(Invalid):
    pass


class JobArgExists(Exists):
    pass


class JobArgForbiddenCreate(ForbiddenCreate):
    pass


class JobArgForbiddenRead(ForbiddenRead):
    pass


class JobArgForbiddenUpdate(ForbiddenUpdate):
    pass


class JobArgForbiddenDelete(ForbiddenDelete):
    pass


class JobArgUnsupported(Unsupported):
    pass


class NomadAttempt_Invalid(Invalid):
    pass


class NomadAttempt_NotFound(NotFound):
    pass


class NomadAttempt_Exists(Exists):
    pass


class NomadAttempt_ForbiddenCreate(ForbiddenCreate):
    pass


class NomadAttempt_ForbiddenRead(ForbiddenRead):
    pass


class NomadAttempt_ForbiddenUpdate(ForbiddenUpdate):
    pass


class NomadAttempt_ForbiddenDelete(ForbiddenDelete):
    pass


class NomadAttempt_Unsupported(Unsupported):
    pass


class NomadEvent_Invalid(Invalid):
    pass


class NomadEvent_NotFound(NotFound):
    pass


class NomadEvent_Exists(Exists):
    pass


class NomadEvent_ForbiddenCreate(ForbiddenCreate):
    pass


class NomadEvent_ForbiddenRead(ForbiddenRead):
    pass


class NomadEvent_ForbiddenUpdate(ForbiddenUpdate):
    pass


class NomadEvent_ForbiddenDelete(ForbiddenDelete):
    pass


class NomadEvent_Unsupported(Unsupported):
    pass


class Notification_Invalid(Invalid):
    pass


class Notification_NotFound(NotFound):
    pass


class Notification_Exists(Exists):
    pass


class NotificationNotFound(NotFound):
    pass


class NotificationInvalid(Invalid):
    pass


class NotificationExists(Exists):
    pass


class NotificationForbiddenCreate(ForbiddenCreate):
    pass


class NotificationForbiddenRead(ForbiddenRead):
    pass


class NotificationForbiddenUpdate(ForbiddenUpdate):
    pass


class NotificationForbiddenDelete(ForbiddenDelete):
    pass


class NotificationUnsupported(Unsupported):
    pass


class TaskNotFound(NotFound):
    pass


class TaskInvalid(Invalid):
    pass


class TaskExists(Exists):
    pass


class TaskForbiddenCreate(ForbiddenCreate):
    pass


class TaskForbiddenRead(ForbiddenRead):
    pass


class TaskForbiddenUpdate(ForbiddenUpdate):
    pass


class TaskForbiddenDelete(ForbiddenDelete):
    pass


class TaskUnsupported(Unsupported):
    pass


class ActionNotFound(NotFound):
    pass


class ActionInvalid(Invalid):
    pass


class ActionExists(Exists):
    pass


class ActionForbiddenCreate(ForbiddenCreate):
    pass


class ActionForbiddenRead(ForbiddenRead):
    pass


class ActionForbiddenUpdate(ForbiddenUpdate):
    pass


class ActionForbiddenDelete(ForbiddenDelete):
    pass


class ActionUnsupported(Unsupported):
    pass


class UserNotFound(NotFound):
    pass


class UserInvalid(Invalid):
    pass


class UserExists(Exists):
    pass


class UserForbiddenCreate(ForbiddenCreate):
    pass


class UserForbiddenRead(ForbiddenRead):
    pass


class UserForbiddenUpdate(ForbiddenUpdate):
    pass


class UserForbiddenDelete(ForbiddenDelete):
    pass


class UserUnsupported(Unsupported):
    pass


class SessionNotFound(NotFound):
    pass


class SessionInvalid(Invalid):
    pass


class SessionExists(Exists):
    pass


class SessionForbiddenCreate(ForbiddenCreate):
    pass


class SessionForbiddenRead(ForbiddenRead):
    pass


class SessionForbiddenUpdate(ForbiddenUpdate):
    pass


class SessionForbiddenDelete(ForbiddenDelete):
    pass


class SessionUnsupported(Unsupported):
    pass


class DatasetNotFound(NotFound):
    pass


class DatasetInvalid(Invalid):
    pass


class DatasetExists(Exists):
    pass


class DatasetForbiddenCreate(ForbiddenCreate):
    pass


class DatasetForbiddenRead(ForbiddenRead):
    pass


class DatasetForbiddenUpdate(ForbiddenUpdate):
    pass


class DatasetForbiddenDelete(ForbiddenDelete):
    pass


class DatasetUnsupported(Unsupported):
    pass


class SecretNotFound(NotFound):
    pass


class SecretInvalid(Invalid):
    pass


class SecretExists(Exists):
    pass


class SecretForbiddenCreate(ForbiddenCreate):
    pass


class SecretForbiddenRead(ForbiddenRead):
    pass


class SecretForbiddenUpdate(ForbiddenUpdate):
    pass


class SecretForbiddenDelete(ForbiddenDelete):
    pass


class SecretUnsupported(Unsupported):
    pass


PRODIGY_TEAMS_ERRORS = {
    k: v
    for k, v in globals().items()
    if type(v) is type and issubclass(v, ProdigyTeamsError)
}



from ..models import (
    BrokerPathCreating,
    BrokerPathDeleting,
    BrokerPathReading,
    BrokerPathReturning,
    BrokerPathUpdating,
)
from .base import ModelClient


class BrokerPath(
    ModelClient[
        BrokerPathCreating,
        BrokerPathReading,
        BrokerPathUpdating,
        BrokerPathDeleting,
        BrokerPathReturning,
        BrokerPathReturning,
    ]
):
    Creating = BrokerPathCreating
    Reading = BrokerPathReading
    Updating = BrokerPathUpdating
    Deleting = BrokerPathDeleting
    Returning = BrokerPathReturning



from radicli import Arg
from wasabi import msg

from .. import ty
from ..auth import get_current_auth
from ..cli import cli
from ..errors import BrokerError, CLIError
from ..messages import Messages
from ..prodigy_teams_broker_sdk.models import Secret, SecretResponse


@cli.subcommand(
    "consul",
    "create",
    key=Arg(help=Messages.path.format(noun="key")),
    value=Arg(help=Messages.value.format(noun="key")),
)
def create(key: str, value: str) -> None:
    """Create or update secret key and value"""
    auth = get_current_auth()
    try:
        auth.broker_client.consul.create(Secret(key=key, value=value))
    except BrokerError as e:
        raise CLIError(Messages.E001.format(noun="secret", name=key), e)
    msg.good(Messages.T002.format(noun="secret", name=key))


@cli.subcommand(
    "consul",
    "info",
    key=Arg(help=Messages.path.format(noun="key")),
)
def info(key: str) -> ty.Optional[ty.List[SecretResponse]]:
    """Get information related to secret key"""
    auth = get_current_auth()
    try:
        res = auth.broker_client.consul.read(key=key)
    except BrokerError as e:
        raise CLIError(Messages.E008.format(noun="secret", name=key), e)
    msg.table({o.Key: o.Value for o in res}, header=["key", "value"])
    return res


@cli.subcommand(
    "consul",
    "delete",
    key=Arg(help=Messages.path.format(noun="key")),
)
def delete(key: str) -> bool:
    """Delete a secret from a broker cluster"""
    auth = get_current_auth()
    try:
        res = auth.broker_client.consul.delete(key=key)
    except BrokerError as e:
        raise CLIError(Messages.E005.format(noun="secret", name=key), e)
    msg.good(Messages.T003.format(noun="secret", name=key))
    return res
